#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/ipc.h>
#include <sys/shm.h>

const char *decryption_key = NULL;
const char chosen_plaintext[] = "ABCDEFGHIJKLMNOPQRSTUVWXYZ012345";  // needs to be same length as the key
const int plaintext_repetitions = 3;

void handle(const char *ptr, size_t len)
{
	if (decryption_key != NULL) {
		int i;
		for (i = 0; i < len; ++i) {
			putchar(ptr[i] ^ decryption_key[i % strlen(decryption_key)]);
		}
		printf("\n");
		return;
	}

	printf("Encrypted record of length %d\n", len);
	if (len >= strlen(chosen_plaintext) * plaintext_repetitions) {
		int i, j;
		printf("Long enough to try chosen-plaintext attack...\n");
		for (i = 0; i <= len - strlen(chosen_plaintext) * plaintext_repetitions; ++i) {
			int num_match = 1;
			for (j = 1; j < plaintext_repetitions; ++j) {
				if (memcmp(ptr + i, ptr + i + j * strlen(chosen_plaintext), strlen(chosen_plaintext)) == 0) {
					++num_match;
				}
			}
			if (num_match == plaintext_repetitions) {
				char buf[256];
				for (j = 0; j < strlen(chosen_plaintext); ++j) {
					buf[(i + j) % strlen(chosen_plaintext)] = ptr[i + j] ^ chosen_plaintext[j];
				}
				buf[strlen(chosen_plaintext)] = 0;
				printf("Plaintext attack at offset %d succeeded! Candidate key: '%s'\n", i, buf);
			}
		}
	}
}

void decrypt(const char *ptr, size_t len)
{
	size_t v7;
	unsigned int* v16 = 0;
	unsigned int* v17;
	signed int v18;
	char dest[4096];

	v16 = (unsigned int*)ptr;
	if (len == 0 || *v16 == 0) {
		return;
	}

	v17 = (unsigned int *)(ptr + 8);
	v18 = 1;
	for ( ;; ) {
		int record_len = *v17;
		memcpy(dest, (const char *)v17 + 4, record_len);
		v7 = record_len;
		handle(dest, record_len);
		++v18;
		if (*v16 < (unsigned int)v18)
			break;
		v17 = (unsigned int *)((char*)v17 + 4 + record_len);
	}
}

void dump(key_t key, size_t size)
{
	FILE *fp;
	int shm;
	char filename[256];
	void *buf;
	
	if (decryption_key == NULL) {
		printf("No decryption key given. Will attempt chosen-plaintext attack (see README).\n");
	} else {
		printf("Trying to decrypt using key '%s'\n", decryption_key);
	}

 	shm = shmget(key, size, 0);
	if (shm == -1) {
		perror("shmget");
		exit(1);
	}

	buf = shmat(shm, NULL, 0);
	if (buf == NULL) {
		perror("shmat");
		exit(1);
	}

	sprintf(filename, "dump.%04x", key);
	fp = fopen(filename, "wb");
	fwrite(buf, size, 1, fp);
	fclose(fp);

	printf("Dumped segment to %s\n", filename);

	decrypt(buf, size);

	shmdt(buf);
	close(shm);
}

int main(int argc, char **argv)
{
	int i;
	int found = 0;

	decryption_key = argv[1]; 

	for (i = 0; i < 10000; ++i) {
		struct shmid_ds buf;
		int ret = shmctl(i, SHM_STAT, &buf);
		if (ret == -1) {
			continue;
		}
		if (buf.shm_perm.__key < 0x10000 &&
		    (buf.shm_segsz == 2097220 ||
		     buf.shm_segsz == 2097224) &&
		    buf.shm_perm.uid == 0 &&
		    buf.shm_perm.mode == 0666) {
			printf("\nSuspicious segment with key 0x%04x found!\n", buf.shm_perm.__key);
			dump(buf.shm_perm.__key, buf.shm_segsz);
			++found;
		}
		++i;
	}

	if (found == 0) {
		printf("No suspicious segments found.\n");
	}

	exit(0);
}

